--[[----------------------------------------------------------------------------

WFCatSpaceOzTestSession.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local KSDate = require 'KSDate'
local KSIdGenerator = require 'KSIdGenerator'
local KSJsonEncoder = require 'KSJsonEncoder'
local WFArguments = require 'WFArguments'

	-- WF.managed_catalog
local WFCatCollaboration = require 'WFCatCollaboration'
local WFCatTestSession = require 'WFCatTestSession'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--==============================================================================

local WFCatSpaceOzTestSession = {}

--------------------------------------------------------------------------------

function WFCatSpaceOzTestSession.setupOzTest( rxTest, memberRole, enableFeature, uploadOriginalBinary )

	local indexDefinitions
	
	if require 'WFPlatformInfo'.isMobile() then
		indexDefinitions = 'LrmCatalogIndexDefinitions.lua'
	end

	local ownerCatalogTestSession = WFCatTestSession( rxTest, nil, indexDefinitions )

	-- Create the owner's catalog.

	local ownerWFCatalog, ownerCatalogSpec

	rxTest:waitFor(

		ownerCatalogTestSession:catalog()
			:map( function( _catalog )
				check( 'WFCatalog' )
				ownerWFCatalog = _catalog
			end ) )

	rxTest:waitFor(

		ownerCatalogTestSession:catalogSpec()
			:tap( function( _catalogSpec )
				check( 'WFCatSpec' )
				ownerCatalogSpec = _catalogSpec
			end ) )
		
	if enableFeature then
	
		rxTest:waitFor( ownerWFCatalog:ozUser():enableFeature( enableFeature ) )
		rxTest:waitFor( ownerWFCatalog:ozUser():getAccountDocFromOz() )
	
	end
		
	-- Create a shared album. Add an album with an asset to the catalog.
	
	-- asset document ...

	local assetId = KSIdGenerator.uuidInLowercaseHex()
	local assetDocRevId = KSIdGenerator.uuidInLowercaseHex()

	local assetDocPutUrl = string.format( '%s/assets/%s/revisions/%s',
										  ownerCatalogSpec:catalogHref(),
										  assetId, assetDocRevId )

	local assetDocContent = KSJsonEncoder.encodeOzUploadJson {
		user_id = ownerCatalogSpec:wfOzUser():userId(),
		payload = {
			captureDate = "2004-11-02T11:10:09",
			importSource = {
				importTimestamp = '2016-01-29T15:53:04Z',
				importedBy = ownerCatalogSpec:ozUserId(),
				fileName = "fakeFile01.jpg",
				originalHeight = 16,
				originalWidth = 16,
			},
		},
		type = 'asset',
		subtype = 'image',
	}

	local assetDocPutArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
	}

	rxTest:expectOnlyCompleted(
		ownerCatalogSpec:put( assetDocPutUrl, assetDocContent, assetDocPutArgs )
			:filter( function() return false end ) )
			
	-- upload master binary
	
	if uploadOriginalBinary then
	
		local assetOriginalPutBinaryUrl = string.format( '%s/assets/%s/revisions/%s/master',
										  ownerCatalogSpec:catalogHref(),
										  assetId, assetDocRevId )
		
		local originalMock = 'original_data'
		
		local assetOriginalPutBinaryArgs = {
			requestHeaders = {
				{ header = 'X-Generate-Renditions', value = 'all' },
				{ header = 'Content-Type', value = 'application/octet-stream' },
				{ header = "Accept", value = "application/json" },
			},
		}
		
		rxTest:waitFor( ownerCatalogSpec:putBinary( assetOriginalPutBinaryUrl, originalMock, assetOriginalPutBinaryArgs ) )
	
	end 

	-- album doc ...

	local albumId = KSIdGenerator.uuidInLowercaseHex()

	local albumDocPutUrl = string.format( '%s/albums/%s',
										  ownerCatalogSpec:catalogHref(),
										  albumId )

	local albumDocContent = KSJsonEncoder.encodeOzUploadJson {
		subtype = 'collection',
		payload = {
			name = "WFCatSpaceOzTestSession album",
			
		},
	}

	local albumDocPutArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
	}

	rxTest:expectOnlyCompleted(
		ownerCatalogSpec:put( albumDocPutUrl, albumDocContent, albumDocPutArgs )
			:filter( function() return false end ) )

	-- album_asset doc ...

	local albumAssetId = KSIdGenerator.uuidInLowercaseHex()

	local albumAssetDocPutUrl = string.format( '%s/albums/%s/assets',
											  ownerCatalogSpec:catalogHref(),
											  albumId )

	local albumAssetDocContent = KSJsonEncoder.encodeOzUploadJson {
		resources = {
			{
				id = assetId,
				album_asset_id = albumAssetId,
			},
		},
	}

	local albumAssetDocPutArgs = {
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
	}

	rxTest:expectOnlyCompleted(
		ownerCatalogSpec:put( albumAssetDocPutUrl, albumAssetDocContent, albumAssetDocPutArgs )
			:filter( function() return false end ) )
	
	-- Create the shared space.
	
	local attributes = {
	
		private = true,
		allowInviteRequests = true,
	
		download = false,
		metadata = false,
		location = false,
		pickedOnly = false,
		ratingAtLeast = 0,
		comments_disabled = false,
		favorites_disabled = false,
		hide_url = false,
		
		author = "Edvard Munch",
		title = "The Scream",
		description = "Anyone can look at The Scream and feel something",
		showTitle = true,
		showAuthor = true,
		showCaptions = true,
		showCover = true,
		themeId = "Single",
		styleId = "Light",
		presentationFilters = true,
		
	}
	
	local spaceId
	
	rxTest:expectSingleValue( WFCatCollaboration.shareAlbum( ownerWFCatalog, albumId, attributes )
		:map( function( shareAttributes )
		
			spaceId = shareAttributes.id
			shareAttributes.id = nil
			shareAttributes.url = nil
			return shareAttributes
			
		end ),
		attributes )
	
	-- Verify the space exists and is synced.
	
	rxTest:expectSingleValue( ownerWFCatalog:docStore()
		:docValue( spaceId )
		:map( function( docValue )
		
			return not not docValue._serverProvidedMetadata
		
		end ), true )
	
	-- Now that a shared album/space exists that allow invites, create the member!
	
	local memberOptions = {
		testName = 'spaceFeedManagerMember1'
			-- Specify a custom test name here; this allows us to fetch a second
			-- account from the Oz account server. Otherwise, the account server
			-- will return the same account we previoulsy fetched for this
			-- same test case.
	}
	
	local memberCatalogTestSession = WFCatTestSession( rxTest, nil, indexDefinitions, memberOptions )

	local memberWFCatalog, memberCatalogSpec

	rxTest:waitFor(

		memberCatalogTestSession:catalog()
			:tap( function( _catalog )
				check( 'WFCatalog' )
				memberWFCatalog = _catalog
			end ) )
	
	rxTest:waitFor(

		memberCatalogTestSession:catalogSpec()
			:tap( function( _catalogSpec )
				check( 'WFCatSpec' )
				memberCatalogSpec = _catalogSpec
			end ) )
	
	-- Add the member to the album.
	
	rxTest:expectOnlyCompleted( ownerCatalogTestSession:reloadAccessAndRefreshTokens() )
		-- Make sure we're using the album owners access tokens for the following.
		
	local newSpaceUserId = KSIdGenerator.uuidInLowercaseHex()
	
	local requestAccessUrl = string.format( '%s/spaces/%s/users/%s',
											  ownerCatalogSpec:catalogHref(),
											  spaceId,
											  newSpaceUserId )

	local requestAccessContent = KSJsonEncoder.encodeOzUploadJson {
		account_id = memberWFCatalog:ozUserId(),
		invite_id = KSIdGenerator.uuidInLowercaseHex(),
		payload = {
			status = 'accepted',
			role = memberRole,
			userCreated = KSDate.timeInIso8601Format(),
			userUpdated= KSDate.timeInIso8601Format()
		}
	}

	local requestAccessArgs = {
		requestMethod = 'PUT',
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
			{ header = 'Accept', value = 'application/json' },
		},
	}

	rxTest:expectSingleValue(
		ownerCatalogSpec:put( requestAccessUrl, requestAccessContent, requestAccessArgs )
			:map( function( data, headers )
			
				return headers.status_code == '201'
				
			end ), true )
	
	rxTest:expectOnlyCompleted( memberCatalogTestSession:reloadAccessAndRefreshTokens() )
		-- reload the member access token.
		
	return ownerWFCatalog, ownerCatalogSpec, ownerCatalogTestSession,
			memberWFCatalog, memberCatalogSpec, memberCatalogTestSession,
			spaceId, newSpaceUserId, albumId, assetId

end

--------------------------------------------------------------------------------

return WFCatSpaceOzTestSession
