--[[----------------------------------------------------------------------------

WFRxIndexStoreSchemaVersionList.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2020 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

------------------------------------------------------------------------------]]

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'


local WFArguments_check = assert( WFArguments.check )

--============================================================================--

-- @mark WFRxIndexStoreSchemaVersionList namespace

local WFRxIndexStoreSchemaVersionList = {}

--------------------------------------------------------------------------------

-- The indexStore schema version list definition.
-- sequence: a sequential list of schema versions. The version list based reindexer will
--           reindex all documents which are defined by sequences > the current sequence
--           If the current schema is not in list, fall back to hash based reindexing,
--           sequence = 0, if in this list, simply disables the fall back to hash based reindexing.
-- platforms: list of platforms affected by this sequence. 
--            allowed: iOS, Android, MacOSX, Windows and the composites Desktop, Mobile, All
--            Note: the lookup is case sensitive!
-- tables:    list of tables affected
-- types:     list of document { types, subtype } tuples to be reindexed.
--            Use subtype '*' for all subtypes, 'NULL' for types without subtype

-- Example:
--	{
--		{
--			sequence = 0,
--		},
--		{
--			sequence = 1,
--			platforms = { "Mobile", },
--			tables = { "coreInfo", "album", "album_asset" },
--          types = { { "asset", "video" }, { "album", "*"}, { "album_asset", "NULL" } }
--		}
--	}

--==============================================================================


local indexStoreSchemaVersionList = {
	{
		sequence = 0,
	},
}

--==============================================================================

-- Note: this list contains far more docTypes than what is actually ever indexed in the indexStore. Still these are
-- the docTypes which we could encounter in an actual catalog and we might want to query for them one day.
local validDocTypesAndSubtypes = {
		catalog = { "lightroom", "X-lightroom", "profile", "pivot", "temp" }, --  X-lightroom is a documented doc type

		space = { "shared", "adhoc", "gallery", "review" },

		asset = { "image", "video", "creation", "creation_segment", "camera_profile",
			"lens_profile", "layout_segment", "deleted_image", "deleted_video",
			"search", "preset", "profile", "task", "person", "pivot", "stack", "model" }, -- model as subtype is tentative, not yet finalized

		[ "error" ] = { "Undefined", "UnknownError", "ResourceNotFoundError", "ResourceExistsError", "RevisionExistsError",
			"NotHeadRevisionPreconditionError", "InputValidationError", "PayloadValidationError",
			"PreconditionFailed", "InlineXmpDevelopSettingTooBigError", "MultipartCompleteError",
			"ParentSubtypeNotAllowed", "ResourceSubTypeNotAllowed", "AddAssetToAlbumError", "JoinIdAlreadyInUseError",
			"ResourceLockedError", "DocumentTooBigError", "PendingAccessRequestError", "NoAccessRequestError",
			"DuplicateAccessRequestError", "InvalidAccessRequestError", "InsufficientPermissionError", "SpaceUserExistsError",
			"AssetGoneError", "AssetNotRestored", "PurgeError", "RestoreAssetError", "RedisServerError", "IMSAdminProfileRequestError",
			"PipelineError", "StorageExceededError", "IneligibleForEnterpriseError", "UserRemovedError", "TrialOrSubscriptionExpiredError",
			"EntitlementExpiredError", "AuthorizationError", "PermissionError", "TooManyResourcesError", "TooManyAssetsError",
			"TooManyRevisionsError", "SharedCloudError" },

		warning = { "RevisionCompressed" },

		space_album = {},

		album = { "adhoc", "cluster", "collection", "collection_set", "export", "project", "project_set", "topic" },

		album_asset = {},

		stack_asset = {},

		account = {},

		comment = { "user" },

		favorite = { "user" },

		feedbackuser = {},

		spaceinvite = { "request", "single_use" },

		space_user = {},

		version = { "edit", "version" }, -- subtype version: undocumented, but found in one catalog, early version implementation?

		watermark = { "text" },
		
		-- client specific docTypes
		[ "gen-album" ] = {}, -- LriOS generic album type

		profile = {}, -- Found in old QE LrDesktop catalog. Subtype "camera" is written in _localOnly. Probably early camera profile implementation or bug.

}

-------------------------------------------------------------------------------

local validPlatforms = {
	iOS = true,
	Android = true,
	MacOSX = true,
	Windows = true,
	Desktop = true,
	Mobile = true,
	All = true,
}

-------------------------------------------------------------------------------

local function sequenceAffectsThisPlatform( platforms )

	local platform = WFPlatformInfo.getOS()

	for _, v in ipairs( platforms ) do
	
		if validPlatforms[ v ] == nil then
			error( string.format( "WFRxIndexStoreSchemaVersionList: invalid platform specificer: %s", v ) )
		end

		if v == platform then
			return true
		end

		if WFPlatformInfo.isMobile() and v == "Mobile" then
			return true
		end

		if WFPlatformInfo.isDesktop and v == "Desktop" then
			return true
		end
		if v == "All" then
			return true
		end
	end

	return false
end

-------------------------------------------------------------------------------

local function isInList( t, list )
	
	for _, v in ipairs( list ) do
		if v == t then 
			return true
		end
	end

	return false

end

-------------------------------------------------------------------------------

local function addTables( tables, toBeReindexedTables, validTables )

	for _, t in ipairs( tables ) do

		if not isInList( t, validTables ) then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTables(): invalid table '%s' for this platform", t) )
		end

		if not isInList( t, toBeReindexedTables ) then
			toBeReindexedTables[ #toBeReindexedTables +1 ] = t
		end

	end

end

-------------------------------------------------------------------------------
	
local function addTypes( types, toBeReindexedTypes )

	for _, typeAndSubtype in ipairs( types ) do

		if type( typeAndSubtype ) ~= 'table' and #typeAndSubtype ~= 2 then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTypes(): invalid types entry in schema version list '%s'", typeAndSubtype) )
		end

		local docType, docSubtype = unpack( typeAndSubtype )

		if validDocTypesAndSubtypes[ docType ] == nil then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTypes(): invalid doc type '%s'", docType) )
		end
		local validSubtypes = validDocTypesAndSubtypes[ docType ]

		if docSubtype == "NULL" and #validSubtypes > 0 then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTypes(): docs of type '%s' require a subtype", docType) )
		end
		if docSubtype == "*" and #validSubtypes == 0 then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTypes(): docs of type '%s' have no subtypes", docType) )
		end

		if docSubtype ~= "NULL" and docSubtype ~= "*" and not isInList( docSubtype, validSubtypes ) then
			error( string.format( "WFRxIndexStoreSchemaVersionList.addTypes(): invalid subtype '%s' to doc type: '%s'", 
				docSubtype, docType) )
		end

		if docSubtype == "*" then
			toBeReindexedTypes[ docType ] = { "*" } -- it's OK to overwrite an already existing subtype list
		elseif docSubtype == "NULL" then
			toBeReindexedTypes[ docType ] = { "NULL" } -- it's OK to overwrite an already existing subtype list
		else
			local alreadyAddedSubtypes = toBeReindexedTypes[ docType ] or {}
			if not isInList( "*", alreadyAddedSubtypes ) and not isInList( docSubtype, alreadyAddedSubtypes )  then
				alreadyAddedSubtypes[ #alreadyAddedSubtypes + 1 ] = docSubtype
				toBeReindexedTypes[ docType ] = alreadyAddedSubtypes
			end
		end
	end

end

--------------------------------------------------------------------------------

function WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables( currentSequence, validTables )

	WFArguments_check( 'number', 'table')

	if #indexStoreSchemaVersionList == 0 then
		error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): invalid schema version list")
	end
	
	local lastSequence = -1
	
	local toBeReindexedTables = {}
	local toBeReindexedTypes = {}
	
	for _, v in ipairs( indexStoreSchemaVersionList ) do
		
		if type( v ) ~= 'table' then
			error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): invalid schema version list, expected table" )
		end

		if type( v.sequence ) ~= 'number' or v.sequence <= lastSequence then
			error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): invalid sequence" )
		end
		
		lastSequence = v.sequence
		
		if v.sequence > 0 then
		
			if type( v.platforms ) ~= 'table' or #v.platforms == 0 then
				error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): invalid platforms" )
			end

			if type( v.tables) ~= 'table' or #v.tables == 0 then
				error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): tables must be a list" )
			end
		
			if type( v.types) ~= 'table' or #v.platforms == 0 then
				error( "WFRxIndexStoreSchemaVersionList.getReindexTypesAndTables(): types must be a list" )
			end

			if v.sequence > currentSequence then 
				if sequenceAffectsThisPlatform( v.platforms ) then
					addTables( v.tables, toBeReindexedTables, validTables )
					addTypes( v.types, toBeReindexedTypes )
				end
			end

		end

	end
		
	local baseSequence = indexStoreSchemaVersionList[ 1 ].sequence
	
	-- currentSequence is no longer in list, fall back to hash based reindexing
	if  baseSequence > currentSequence and baseSequence ~= 0 then
		return lastSequence, "hashBasedIndexing"
	else
		return lastSequence, toBeReindexedTypes, toBeReindexedTables
	end

end

--------------------------------------------------------------------------------

function WFRxIndexStoreSchemaVersionList.TEST_setSchemaVersionList( schemaVersionList )

	WFArguments_check( 'table')
	
	indexStoreSchemaVersionList = schemaVersionList

end

--------------------------------------------------------------------------------

return WFRxIndexStoreSchemaVersionList

